@@ -153,39 +153,59 @@ bus_to_str (WacomBusType bus)
 }
 
 char *
-make_match_string (const char *name, WacomBusType bus, int vendor_id, int product_id)
+make_match_string (const char *name, const char *uniq, WacomBusType bus, int vendor_id, int product_id)
 {
-	return g_strdup_printf("%s:%04x:%04x%s%s",
+	return g_strdup_printf("%s|%04x|%04x%s%s%s%s",
 				bus_to_str (bus),
 				vendor_id, product_id,
-				name ? ":" : "",
-				name ? name : "");
+				(name || uniq) ? "|" : "",
+				name ? name : "",
+			        uniq ? "|" : "",
+			        uniq ? uniq : "");
 }
 
 static gboolean
-match_from_string(const char *str, WacomBusType *bus, int *vendor_id, int *product_id, char **name)
+match_from_string(const char *str_in, WacomBusType *bus, int *vendor_id, int *product_id, char **name, char **uniq)
 {
-	int rc = 1, len = 0;
-	char busstr[64];
+	gboolean rc = FALSE;
+	guint64 num;
+	char *str = g_strdup(str_in);
+	char **components = NULL;
 
-	rc = sscanf(str, "%63[^:]:%x:%x:%n", busstr, vendor_id, product_id, &len);
-	if (len > 0) {
-		/* Grumble grumble scanf handling of %n */
-		*name = g_strdup(str+len);
-	} else if (rc == 3) {
-		*name = NULL;
-	} else {
-		return FALSE;
+	if (g_str_has_suffix(str_in, ";"))
+		str[strlen(str) - 1] = '\0';
+
+	components = g_strsplit(str, "|", 16);
+	if (!components[0] || !components[1] || !components[2])
+		goto out;
+
+	*bus = bus_from_str (components[0]);
+	if (!g_ascii_string_to_unsigned(components[1], 16, 0, 0xffff, &num, NULL))
+		goto out;
+	*vendor_id = (int)num;
+
+	if (!g_ascii_string_to_unsigned(components[2], 16, 0, 0xffff, &num, NULL))
+		goto out;
+	*product_id = (int)num;
+
+	if (components[3]) {
+		*name = g_strdup(components[3]);
+		if (components[4])
+			*uniq = g_strdup(components[4]);
 	}
-	*bus = bus_from_str (busstr);
 
-	return TRUE;
+	rc = TRUE;
+out:
+	free(str);
+	g_strfreev(components);
+	return rc;
 }
 
 static WacomMatch *
 libwacom_match_from_string(const char *matchstr)
 {
 	char *name = NULL;
+	char *uniq = NULL;
 	int vendor_id, product_id;
 	WacomBusType bus;
 	WacomMatch *match;
@@ -195,16 +215,18 @@ libwacom_match_from_string(const char *matchstr)
 
 	if (g_str_equal(matchstr, GENERIC_DEVICE_MATCH)) {
 		name = NULL;
+		uniq = NULL;
 		bus = WBUSTYPE_UNKNOWN;
 		vendor_id = 0;
 		product_id = 0;
-	} else if (!match_from_string(matchstr, &bus, &vendor_id, &product_id, &name)) {
+	} else if (!match_from_string(matchstr, &bus, &vendor_id, &product_id, &name, &uniq)) {
 		DBG("failed to match '%s' for product/vendor IDs. Skipping.\n", matchstr);
 		return NULL;
 	}
 
-	match = libwacom_match_new(name, bus, vendor_id, product_id);
+	match = libwacom_match_new(name, uniq, bus, vendor_id, product_id);
 	free(name);
+	free(uniq);
 
 	return match;
 }
@@ -213,22 +235,50 @@ static gboolean
 libwacom_matchstr_to_paired(WacomDevice *device, const char *matchstr)
 {
 	char *name = NULL;
+	char *uniq = NULL;
 	int vendor_id, product_id;
 	WacomBusType bus;
 
 	g_return_val_if_fail(device->paired == NULL, FALSE);
 
-	if (!match_from_string(matchstr, &bus, &vendor_id, &product_id, &name)) {
+	if (!match_from_string(matchstr, &bus, &vendor_id, &product_id, &name, &uniq)) {
 		DBG("failed to match '%s' for product/vendor IDs. Ignoring.\n", matchstr);
 		return FALSE;
 	}
 
-	device->paired = libwacom_match_new(name, bus, vendor_id, product_id);
+	device->paired = libwacom_match_new(name, uniq, bus, vendor_id, product_id);
 
 	free(name);
+	free(uniq);
 	return TRUE;
 }
 
+static bool
+parse_stylus_id(const char *str, WacomStylusId *id)
+{
+	char **tokens = g_strsplit(str, ":", 2);
+	const char *vidstr, *tidstr;
+	int vid, tool_id;
+	bool ret = false;
+
+	if (tokens[1] == NULL) {
+		vidstr = G_STRINGIFY(WACOM_VENDOR_ID);
+		tidstr = tokens[0];
+	} else {
+		vidstr = tokens[0];
+		tidstr = tokens[1];
+	}
+
+	if (safe_atoi_base (vidstr, &vid, 16) && safe_atoi_base (tidstr, &tool_id, 16)) {
+		id->vid = vid;
+		id->tool_id = tool_id;
+		ret = true;
+	}
+	g_clear_pointer(&tokens, g_strfreev);
+
+	return ret;
+}
+
 static void
 libwacom_parse_stylus_keyfile(WacomDeviceDatabase *db, const char *path)
 {
@@ -242,46 +292,46 @@ libwacom_parse_stylus_keyfile(WacomDeviceDatabase *db, const char *path)
 
 	rc = g_key_file_load_from_file(keyfile, path, G_KEY_FILE_NONE, &error);
 	g_assert (rc);
+
 	groups = g_key_file_get_groups (keyfile, NULL);
 	for (i = 0; groups[i]; i++) {
 		WacomStylus *stylus;
+		WacomStylusId id;
 		GError *error = NULL;
-		char *type;
-		int id;
+		char *eraser_type, *type;
 		char **string_list;
 
-		if (!safe_atoi_base (groups[i], &id, 16)) {
-			g_warning ("Failed to parse stylus ID '%s'", groups[i]);
+		if (!parse_stylus_id(groups[i], &id)) {
+			g_warning ("Failed to parse stylus ID '%s', ignoring entry", groups[i]);
 			continue;
 		}
-
 		stylus = g_new0 (WacomStylus, 1);
 		stylus->refcnt = 1;
 		stylus->id = id;
 		stylus->name = g_key_file_get_string(keyfile, groups[i], "Name", NULL);
 		stylus->group = g_key_file_get_string(keyfile, groups[i], "Group", NULL);
+		stylus->paired_stylus_ids = g_array_new (FALSE, FALSE, sizeof(WacomStylusId));
 
-		type = g_key_file_get_string(keyfile, groups[i], "EraserType", NULL);
-		stylus->eraser_type = eraser_type_from_str (type);
-		g_free (type);
+		eraser_type = g_key_file_get_string(keyfile, groups[i], "EraserType", NULL);
+		stylus->eraser_type = eraser_type_from_str (eraser_type);
+		g_clear_pointer(&eraser_type, g_free);
+
+		/* We have to keep the integer array for libwacom_get_supported_styli() */
+		stylus->deprecated_paired_ids = g_array_new (FALSE, FALSE, sizeof(int));
+		stylus->paired_styli = g_array_new (FALSE, FALSE, sizeof(WacomStylus*));
 
 		string_list = g_key_file_get_string_list (keyfile, groups[i], "PairedStylusIds", NULL, NULL);
-		stylus->paired_ids = g_array_new (FALSE, FALSE, sizeof(int));
-		if (string_list) {
-			guint j;
-
-			for (j = 0; string_list[j]; j++) {
-				int val;
-
-				if (safe_atoi_base (string_list[j], &val, 16)) {
-					g_array_append_val (stylus->paired_ids, val);
-				} else {
-					g_warning ("Stylus %s (%s) Ignoring invalid PairedId value\n", stylus->name, groups[i]);
-				}
+		for (guint j = 0; string_list && string_list[j]; j++) {
+			WacomStylusId paired_id;
+			if (parse_stylus_id(string_list[j], &paired_id)) {
+				g_array_append_val (stylus->paired_stylus_ids, paired_id);
+				if (paired_id.vid == WACOM_VENDOR_ID)
+					g_array_append_val (stylus->deprecated_paired_ids, paired_id.tool_id);
+			} else {
+				g_warning ("Stylus %s (%s) Ignoring invalid PairedStylusIds value\n", stylus->name, groups[i]);
 			}
-
-			g_strfreev (string_list);
 		}
+		g_clear_pointer(&string_list, g_strfreev);
 
 		stylus->has_lens = g_key_file_get_boolean(keyfile, groups[i], "HasLens", &error);
 		if (error && error->code == G_KEY_FILE_ERROR_INVALID_VALUE)
@@ -298,49 +348,42 @@ libwacom_parse_stylus_keyfile(WacomDeviceDatabase *db, const char *path)
 			g_clear_error (&error);
 		}
 
+		stylus->axes = WACOM_AXIS_TYPE_NONE;
 		string_list = g_key_file_get_string_list (keyfile, groups[i], "Axes", NULL, NULL);
-		if (string_list) {
-			WacomAxisTypeFlags axes = WACOM_AXIS_TYPE_NONE;
-			guint j;
-
-			for (j = 0; string_list[j]; j++) {
-				WacomAxisTypeFlags flag = WACOM_AXIS_TYPE_NONE;
-				if (g_str_equal(string_list[j], "Tilt")) {
-					flag = WACOM_AXIS_TYPE_TILT;
-				} else if (g_str_equal(string_list[j], "RotationZ")) {
-					flag = WACOM_AXIS_TYPE_ROTATION_Z;
-				} else if (g_str_equal(string_list[j], "Distance")) {
-					flag = WACOM_AXIS_TYPE_DISTANCE;
-				} else if (g_str_equal(string_list[j], "Pressure")) {
-					flag = WACOM_AXIS_TYPE_PRESSURE;
-				} else if (g_str_equal(string_list[j], "Slider")) {
-					flag = WACOM_AXIS_TYPE_SLIDER;
-				} else {
-					g_warning ("Invalid axis %s for stylus ID %s\n",
-						   string_list[j], groups[i]);
-				}
-				if (axes & flag)
-					g_warning ("Duplicate axis %s for stylus ID %s\n",
-						   string_list[j], groups[i]);
-				axes |= flag;
+		for (guint j = 0; string_list && string_list[j]; j++) {
+			WacomAxisTypeFlags flag = WACOM_AXIS_TYPE_NONE;
+			if (g_str_equal(string_list[j], "Tilt")) {
+				flag = WACOM_AXIS_TYPE_TILT;
+			} else if (g_str_equal(string_list[j], "RotationZ")) {
+				flag = WACOM_AXIS_TYPE_ROTATION_Z;
+			} else if (g_str_equal(string_list[j], "Distance")) {
+				flag = WACOM_AXIS_TYPE_DISTANCE;
+			} else if (g_str_equal(string_list[j], "Pressure")) {
+				flag = WACOM_AXIS_TYPE_PRESSURE;
+			} else if (g_str_equal(string_list[j], "Slider")) {
+				flag = WACOM_AXIS_TYPE_SLIDER;
+			} else {
+				g_warning ("Invalid axis %s for stylus ID %s\n",
+					   string_list[j], groups[i]);
 			}
-
-			stylus->axes = axes;
-			g_strfreev (string_list);
+			if (stylus->axes & flag)
+				g_warning ("Duplicate axis %s for stylus ID %s\n",
+					   string_list[j], groups[i]);
+			stylus->axes |= flag;
 		}
+		g_clear_pointer(&string_list, g_strfreev);
 
 		type = g_key_file_get_string(keyfile, groups[i], "Type", NULL);
 		stylus->type = type_from_str (type);
-		g_free (type);
+		g_clear_pointer(&type, g_free);
 
-		if (g_hash_table_lookup (db->stylus_ht, GINT_TO_POINTER (id)) != NULL)
-			g_warning ("Duplicate definition for stylus ID '%#x'", id);
+		if (g_hash_table_lookup (db->stylus_ht, &id) != NULL)
+			g_warning ("Duplicate definition for stylus ID '%s'", groups[i]);
 
-		g_hash_table_insert (db->stylus_ht, GINT_TO_POINTER (id), stylus);
+		g_hash_table_insert (db->stylus_ht, g_memdup2(&id, sizeof(id)), stylus);
 	}
-	g_strfreev (groups);
-	if (keyfile)
-		g_key_file_free (keyfile);
+	g_clear_pointer(&groups, g_strfreev);
+	g_clear_pointer(&keyfile, g_key_file_free);
 }
 
 static void
@@ -352,23 +395,24 @@ libwacom_setup_paired_attributes(WacomDeviceDatabase *db)
 	g_hash_table_iter_init(&iter, db->stylus_ht);
 	while (g_hash_table_iter_next (&iter, &key, &value)) {
 		WacomStylus *stylus = value;
-		const int* ids;
-		int count;
-		int i;
+		GArray *paired_ids = g_steal_pointer(&stylus->paired_stylus_ids);
 
-		ids = libwacom_stylus_get_paired_ids(stylus, &count);
-		for (i = 0; i < count; i++) {
-			const WacomStylus *pair;
+		for (guint i = 0; i < paired_ids->len; i++) {
+			WacomStylusId *id = &g_array_index(paired_ids, WacomStylusId, i);
+			WacomStylus *paired = g_hash_table_lookup(db->stylus_ht, id);
 
-			pair = libwacom_stylus_get_for_id(db, ids[i]);
-			if (pair == NULL) {
-				g_warning("Paired stylus '0x%x' not found, ignoring.", ids[i]);
+			if (paired == NULL) {
+				g_warning ("Ignoring paired stylus %04x:%x", id->vid, id->tool_id);
 				continue;
 			}
-			if (libwacom_stylus_is_eraser(pair)) {
+
+			g_array_append_val(stylus->paired_styli, paired);
+
+			if (libwacom_stylus_is_eraser(paired)) {
 				stylus->has_eraser = true;
 			}
 		}
+		g_array_unref(paired_ids);
 	}
 }
 
@@ -384,7 +428,9 @@ static const struct {
 	{ "Ring2", WACOM_BUTTON_RING2_MODESWITCH },
 	{ "Touchstrip", WACOM_BUTTON_TOUCHSTRIP_MODESWITCH },
 	{ "Touchstrip2", WACOM_BUTTON_TOUCHSTRIP2_MODESWITCH },
-	{ "OLEDs", WACOM_BUTTON_OLED }
+	{ "OLEDs", WACOM_BUTTON_OLED },
+	{ "Dial", WACOM_BUTTON_DIAL_MODESWITCH },
+	{ "Dial2", WACOM_BUTTON_DIAL2_MODESWITCH },
 };
 
 static const struct {
@@ -394,7 +440,9 @@ static const struct {
 	{ "Ring",		WACOM_STATUS_LED_RING },
 	{ "Ring2",		WACOM_STATUS_LED_RING2 },
 	{ "Touchstrip",		WACOM_STATUS_LED_TOUCHSTRIP },
-	{ "Touchstrip2",	WACOM_STATUS_LED_TOUCHSTRIP2 }
+	{ "Touchstrip2",	WACOM_STATUS_LED_TOUCHSTRIP2 },
+	{ "Dial",		WACOM_STATUS_LED_DIAL },
+	{ "Dial2",		WACOM_STATUS_LED_DIAL2 },
 };
 
 static const struct {
@@ -402,7 +450,8 @@ static const struct {
 	WacomIntegrationFlags   value;
 } integration_flags[] = {
 	{ "Display",		WACOM_DEVICE_INTEGRATED_DISPLAY },
-	{ "System",		WACOM_DEVICE_INTEGRATED_SYSTEM }
+	{ "System",		WACOM_DEVICE_INTEGRATED_SYSTEM },
+	{ "Remote",		WACOM_DEVICE_INTEGRATED_REMOTE }
 };
 
 static void
@@ -647,6 +696,8 @@ libwacom_parse_buttons(WacomDevice *device,
 	device->ring_num_modes = libwacom_parse_num_modes(device, keyfile, "RingNumModes", WACOM_BUTTON_RING_MODESWITCH);
 	device->ring2_num_modes = libwacom_parse_num_modes(device, keyfile, "Ring2NumModes", WACOM_BUTTON_RING2_MODESWITCH);
 	device->strips_num_modes = libwacom_parse_num_modes(device, keyfile, "StripsNumModes", WACOM_BUTTON_TOUCHSTRIP_MODESWITCH);
+	device->dial_num_modes = libwacom_parse_num_modes(device, keyfile, "DialNumModes", WACOM_BUTTON_DIAL_MODESWITCH);
+	device->dial2_num_modes = libwacom_parse_num_modes(device, keyfile, "Dial2NumModes", WACOM_BUTTON_DIAL2_MODESWITCH);
 }
 
 static void
@@ -672,12 +723,21 @@ libwacom_parse_keys(WacomDevice *device,
 	libwacom_parse_key_codes(device, keyfile);
 }
 
+static int
+wacom_stylus_id_sort(const WacomStylusId *a, const WacomStylusId *b)
+{
+	if (a->vid == b->vid)
+		return a->tool_id - b->tool_id;
+
+	return a->vid - b->vid;
+}
 
 static int
 styli_id_sort(gconstpointer pa, gconstpointer pb)
 {
-	const int *a = pa, *b = pb;
-	return *a > *b ? 1 : *a == *b ? 0 : -1;
+	const WacomStylus *a = *(WacomStylus**)pa, *b = *(WacomStylus**)pb;
+
+	return wacom_stylus_id_sort(&a->id, &b->id);
 }
 
 static void
@@ -687,17 +747,23 @@ libwacom_parse_styli_list(WacomDeviceDatabase *db, WacomDevice *device,
 	GArray *array;
 	guint i;
 
-	array = g_array_new (FALSE, FALSE, sizeof(int));
-	for (i = 0; ids[i]; i++) {
-		const char *id = ids[i];
+	array = g_array_new (FALSE, FALSE, sizeof(WacomStylus*));
+	for (i = 0; ids && ids[i]; i++) {
+		const char *str = ids[i];
 
-		if (g_str_has_prefix(id, "0x")) {
-			int int_value;
-			if (safe_atoi_base (ids[i], &int_value, 16)) {
-				g_array_append_val (array, int_value);
+		if (g_str_has_prefix(str, "0x")) {
+			WacomStylusId id;
+			if (parse_stylus_id(str, &id)) {
+				WacomStylus *stylus = g_hash_table_lookup(db->stylus_ht, &id);
+				if (stylus)
+					g_array_append_val (array, stylus);
+				else
+					g_warning ("Invalid stylus id for '%s', ignoring stylus", str);
+			} else {
+				g_warning ("Invalid stylus id format for '%s', ignoring stylus", str);
 			}
-		} else if (g_str_has_prefix(id, "@")) {
-			const char *group = &id[1];
+		} else if (g_str_has_prefix(str, "@")) {
+			const char *group = &str[1];
 			GHashTableIter iter;
 			gpointer key, value;
 
@@ -705,17 +771,27 @@ libwacom_parse_styli_list(WacomDeviceDatabase *db, WacomDevice *device,
 			while (g_hash_table_iter_next (&iter, &key, &value)) {
 				WacomStylus *stylus = value;
 				if (stylus->group && g_str_equal(group, stylus->group)) {
-					g_array_append_val (array, stylus->id);
+					g_array_append_val (array, stylus);
 				}
 			}
 		} else {
-			g_warning ("Invalid prefix for '%s'!", id);
+			g_warning ("Invalid prefix for '%s', ignoring stylus", str);
 		}
 	}
 	/* Using groups means we don't get the styli in ascending order.
 	   Sort it so the output is predictable */
 	g_array_sort(array, styli_id_sort);
 	device->styli = array;
+
+	/* The legacy PID-only stylus id list */
+	device->deprecated_styli_ids = g_array_new(FALSE, FALSE, sizeof(int));
+	for (guint i = 0; i < device->styli->len; i++) {
+		WacomStylus *stylus = g_array_index(device->styli, WacomStylus*, i);
+		/* This only ever worked for Wacom styli, so let's keep that behavior */
+		if (stylus->id.vid == 0 || stylus->id.vid == WACOM_VENDOR_ID) {
+			g_array_append_val(device->deprecated_styli_ids, stylus->id.tool_id);
+		}
+	}
 }
 
 static void
@@ -730,12 +806,6 @@ libwacom_parse_features(WacomDevice *device, GKeyFile *keyfile)
 	if (g_key_file_get_boolean(keyfile, FEATURES_GROUP, "Touch", NULL))
 		device->features |= FEATURE_TOUCH;
 
-	if (g_key_file_get_boolean(keyfile, FEATURES_GROUP, "Ring", NULL))
-		device->features |= FEATURE_RING;
-
-	if (g_key_file_get_boolean(keyfile, FEATURES_GROUP, "Ring2", NULL))
-		device->features |= FEATURE_RING2;
-
 	if (g_key_file_get_boolean(keyfile, FEATURES_GROUP, "Reversible", NULL))
 		device->features |= FEATURE_REVERSIBLE;
 
@@ -747,15 +817,13 @@ libwacom_parse_features(WacomDevice *device, GKeyFile *keyfile)
 	    device->features & FEATURE_REVERSIBLE)
 		g_warning ("Tablet '%s' is both reversible and integrated in screen. This is impossible", libwacom_get_match(device));
 
-	if (!(device->features & FEATURE_RING) &&
-	    (device->features & FEATURE_RING2))
-		g_warning ("Tablet '%s' has Ring2 but no Ring. This is impossible", libwacom_get_match(device));
-
 	if (!(device->features & FEATURE_TOUCH) &&
 	    (device->features & FEATURE_TOUCHSWITCH))
 		g_warning ("Tablet '%s' has touch switch but no touch tool. This is impossible", libwacom_get_match(device));
 
+	device->num_rings = g_key_file_get_integer(keyfile, FEATURES_GROUP, "NumRings", NULL);
 	device->num_strips = g_key_file_get_integer(keyfile, FEATURES_GROUP, "NumStrips", NULL);
+	device->num_dials = g_key_file_get_integer(keyfile, FEATURES_GROUP, "NumDials", NULL);
 
 	string_list = g_key_file_get_string_list(keyfile, FEATURES_GROUP, "StatusLEDs", NULL, NULL);
 	if (string_list) {
@@ -863,35 +931,39 @@ libwacom_parse_tablet_keyfile(WacomDeviceDatabase *db,
 				}
 			}
 			if (!found)
-				g_warning ("Unrecognized integration flag '%s'", string_list[i]);
+				g_warning ("Unrecognized integration flag '%s', ignoring flag", string_list[i]);
 		}
 		g_strfreev (string_list);
 	}
 
 	layout = g_key_file_get_string(keyfile, DEVICE_GROUP, "Layout", NULL);
-	if (layout) {
+	if (layout && layout[0] != '\0') {
 		/* For the layout, we store the full path to the SVG layout */
 		device->layout = g_build_filename (datadir, "layouts", layout, NULL);
-		g_free (layout);
 	}
+	g_free (layout);
 
 	class = g_key_file_get_string(keyfile, DEVICE_GROUP, "Class", NULL);
 	device->cls = libwacom_class_string_to_enum(class);
 	g_free(class);
 
 	string_list = g_key_file_get_string_list(keyfile, DEVICE_GROUP, "Styli", NULL, NULL);
-	if (string_list) {
-		libwacom_parse_styli_list(db, device, string_list);
-		g_strfreev (string_list);
-	} else {
-		int fallback_eraser = WACOM_ERASER_FALLBACK_ID;
-		int fallback_stylus = WACOM_STYLUS_FALLBACK_ID;
-		device->styli = g_array_new(FALSE, FALSE, sizeof(int));
-		g_array_append_val(device->styli, fallback_eraser);
-		g_array_append_val(device->styli, fallback_stylus);
+	if (!string_list) {
+		GError *error = NULL;
+		if (g_key_file_get_boolean(keyfile, FEATURES_GROUP, "Stylus", &error) ||
+		    g_error_matches (error, G_KEY_FILE_ERROR, G_KEY_FILE_ERROR_KEY_NOT_FOUND)) {
+			string_list = g_new0(char*, 3);
+			string_list[0] = g_strdup_printf("0x0:0x%x", WACOM_ERASER_FALLBACK_ID);
+			string_list[1] = g_strdup_printf("0x0:0x%x", WACOM_STYLUS_FALLBACK_ID);
+		}
+		if (error)
+			g_error_free(error);
 	}
+	libwacom_parse_styli_list(db, device, string_list);
+	g_strfreev (string_list);
 
 	device->num_strips = g_key_file_get_integer(keyfile, FEATURES_GROUP, "NumStrips", NULL);
+	device->num_dials = g_key_file_get_integer(keyfile, FEATURES_GROUP, "NumDials", NULL);
 	device->buttons = g_hash_table_new_full(g_direct_hash, g_direct_equal,
 						NULL, g_free);
 	device->status_leds = g_array_new (FALSE, FALSE, sizeof(WacomStatusLEDs));
@@ -909,8 +981,10 @@ out:
 		g_key_file_free(keyfile);
 	if (error)
 		g_error_free(error);
-	if (!success)
+	if (!success) {
+		g_warning("Ignoring invalid .tablet file %s", filename);
 		device = libwacom_unref(device);
+	}
 
 	return device;
 }
@@ -943,25 +1017,16 @@ is_stylus_file(const struct dirent *entry)
 }
 
 static bool
-load_tablet_files(WacomDeviceDatabase *db, const char *datadir)
+load_tablet_files(WacomDeviceDatabase *db, GHashTable *parsed_filenames, const char *datadir)
 {
 	DIR *dir;
 	struct dirent *file;
 	bool success = false;
-	GHashTable *keyset = NULL;
 
 	dir = opendir(datadir);
 	if (!dir)
 		return errno == ENOENT; /* non-existing directory is ok */
 
-	/* A set of all matches for duplicate detection. We allow duplicates
-	 * across data directories, but we don't allow for duplicates
-	 * within the same data directory.
-	 */
-	keyset = g_hash_table_new_full (g_str_hash, g_str_equal, free, NULL);
-	if (!keyset)
-		goto out;
-
 	while ((file = readdir(dir))) {
 		WacomDevice *d;
 		guint idx = 0;
@@ -969,6 +1034,11 @@ load_tablet_files(WacomDeviceDatabase *db, const char *datadir)
 		if (!is_tablet_file(file))
 			continue;
 
+		if (g_hash_table_lookup(parsed_filenames, file->d_name))
+			continue;
+
+		g_hash_table_add(parsed_filenames, g_strdup(file->d_name));
+
 		d = libwacom_parse_tablet_keyfile(db, datadir, file->d_name);
 		if (!d)
 			continue;
@@ -985,28 +1055,13 @@ load_tablet_files(WacomDeviceDatabase *db, const char *datadir)
 			const char *matchstr;
 
 			matchstr = libwacom_match_get_match_string(match);
-			/* no duplicate matches allowed within the same
-			 * directory */
-			if (g_hash_table_contains(keyset, matchstr)) {
+			/* no duplicate matches allowed */
+			if (g_hash_table_contains(db->device_ht, matchstr)) {
 				g_critical("Duplicate match of '%s' on device '%s'.",
 					   matchstr, libwacom_get_name(d));
 				goto out;
 			}
-			g_hash_table_add(keyset, g_strdup(matchstr));
-
-			/* We already have an entry for this match in the database,
-			 * that takes precedence. Remove the current match
-			 * from the new tablet - that's fine because we
-			 * haven't exposed the tablet yet.
-			 */
-			if (g_hash_table_lookup(db->device_ht, matchstr)) {
-				libwacom_remove_match(d, match);
-				continue;
-			}
-
-			g_hash_table_insert(db->device_ht,
-					    g_strdup (matchstr),
-					    d);
+			g_hash_table_insert(db->device_ht, g_strdup (matchstr), d);
 			libwacom_ref(d);
 			idx++;
 		}
@@ -1016,8 +1071,6 @@ load_tablet_files(WacomDeviceDatabase *db, const char *datadir)
 	success = true;
 
 out:
-	if (keyset)
-		g_hash_table_destroy(keyset);
 	closedir(dir);
 	return success;
 }
@@ -1054,37 +1107,58 @@ load_stylus_files(WacomDeviceDatabase *db, const char *datadir)
 	return true;
 }
 
+static guint
+stylus_hash(WacomStylusId *id)
+{
+	guint64 full_id = (guint64)id->vid << 32 | id->tool_id;
+	return g_int64_hash(&full_id);
+}
+
+static gboolean
+stylus_compare(WacomStylusId *a, WacomStylusId *b)
+{
+	return wacom_stylus_id_sort(a, b) == 0;
+}
+
 static WacomDeviceDatabase *
-database_new_for_paths (size_t npaths, const char **datadirs)
+database_new_for_paths (char * const *datadirs)
 {
 	WacomDeviceDatabase *db;
-	size_t n;
-	const char **datadir;
+	char * const *datadir;
+	GHashTable *parsed_filenames;
+
+	parsed_filenames = g_hash_table_new_full (g_str_hash, g_str_equal, free, NULL);
+	if (!parsed_filenames)
+		return NULL;
 
 	db = g_new0 (WacomDeviceDatabase, 1);
 	db->device_ht = g_hash_table_new_full (g_str_hash,
 					       g_str_equal,
 					       g_free,
 					       (GDestroyNotify) libwacom_destroy);
-	db->stylus_ht = g_hash_table_new_full (g_direct_hash,
-					       g_direct_equal,
-					       NULL,
+	db->stylus_ht = g_hash_table_new_full ((GHashFunc)stylus_hash,
+					       (GEqualFunc)stylus_compare,
+					       (GDestroyNotify) g_free,
 					       (GDestroyNotify) stylus_destroy);
 
-	for (datadir = datadirs, n = npaths; n--; datadir++) {
+	for (datadir = datadirs; *datadir; datadir++) {
 		if (!load_stylus_files(db, *datadir))
 			goto error;
 	}
 
-	for (datadir = datadirs, n = npaths; n--; datadir++) {
-		if (!load_tablet_files(db, *datadir))
+	for (datadir = datadirs; *datadir;  datadir++) {
+		if (!load_tablet_files(db, parsed_filenames, *datadir))
 			goto error;
 	}
 
+	g_hash_table_unref(parsed_filenames);
+
 	/* If we couldn't load _anything_ then something's wrong */
 	if (g_hash_table_size (db->stylus_ht) == 0 ||
-	    g_hash_table_size (db->device_ht) == 0)
+	    g_hash_table_size (db->device_ht) == 0) {
+		g_warning("Zero tablet or stylus files found in datadirs");
 		goto error;
+	}
 
 	libwacom_setup_paired_attributes(db);
 
@@ -1098,18 +1172,42 @@ error:
 LIBWACOM_EXPORT WacomDeviceDatabase *
 libwacom_database_new_for_path (const char *datadir)
 {
-	return database_new_for_paths(1, &datadir);
+	WacomDeviceDatabase *db;
+	char **paths;
+
+	paths = g_strsplit(datadir, ":", 0);
+	db = database_new_for_paths(paths);
+
+	g_strfreev(paths);
+
+	return db;
 }
 
 LIBWACOM_EXPORT WacomDeviceDatabase *
 libwacom_database_new (void)
 {
-	const char *datadir[] = {
+	WacomDeviceDatabase *db;
+	char *xdgdir = NULL;
+	char *xdg_config_home = g_strdup(g_getenv("XDG_CONFIG_HOME"));
+
+	if (!xdg_config_home)
+		xdg_config_home = g_strdup_printf("%s/.config/", g_get_home_dir());
+
+	xdgdir = g_strdup_printf("%s/libwacom", xdg_config_home);
+
+	char *datadir[] = {
+		xdgdir,
 		ETCDIR,
 		DATADIR,
+		NULL,
 	};
 
-	return database_new_for_paths (2, datadir);
+	db = database_new_for_paths(datadir);
+
+	free(xdgdir);
+	free(xdg_config_home);
+
+	return db;
 }
 
 LIBWACOM_EXPORT void
