@@ -40,22 +40,37 @@
 
 #define GENERIC_DEVICE_MATCH "generic"
 #define WACOM_DEVICE_INTEGRATED_UNSET (WACOM_DEVICE_INTEGRATED_NONE - 1U)
+#define WACOM_VENDOR_ID 0x056a
+
+enum GenericStylus {
+	GENERIC_PEN_WITH_ERASER = 0xfffff,
+        GENERIC_ERASER = 0xffffe,
+        GENERIC_PEN_NO_ERASER = 0xffffd,
+};
 
 enum WacomFeature {
 	FEATURE_STYLUS		= (1 << 0),
 	FEATURE_TOUCH		= (1 << 1),
-	FEATURE_RING		= (1 << 2),
-	FEATURE_RING2		= (1 << 3),
 	FEATURE_REVERSIBLE	= (1 << 4),
 	FEATURE_TOUCHSWITCH	= (1 << 5)
 };
 
+struct _WacomBuilder {
+	char *device_name;
+	char *match_name;
+	char *uniq;
+	WacomBusType bus;
+	uint32_t vendor_id;
+	uint32_t product_id;
+};
+
 /* WARNING: When adding new members to this struct
  * make sure to update libwacom_copy_match() ! */
 struct _WacomMatch {
 	gint refcnt;
 	char *match;
 	char *name;
+	char *uniq;
 	WacomBusType bus;
 	uint32_t vendor_id;
 	uint32_t product_id;
@@ -88,14 +103,21 @@ struct _WacomDevice {
 
 	WacomClass cls;
 	int num_strips;
+	int num_rings;
+	int num_dials;
 	uint32_t features;
 	uint32_t integration_flags;
 
 	int strips_num_modes;
+	int dial_num_modes;
+	int dial2_num_modes;
 	int ring_num_modes;
 	int ring2_num_modes;
 
-	GArray *styli;
+	/* for libwacom_get_supported_styli() */
+	GArray *deprecated_styli_ids; /* int */
+	/* for libwacom_get_styli() */
+	GArray *styli; /* WacomStylus* */
 	GHashTable *buttons; /* 'A' : WacomButton */
 	WacomKeycode keycodes[32];
 	size_t num_keycodes;
@@ -107,14 +129,21 @@ struct _WacomDevice {
 	gint refcnt; /* for the db hashtable */
 };
 
+typedef struct _WacomStylusId {
+	unsigned int vid;
+	unsigned int tool_id;
+} WacomStylusId;
+
 struct _WacomStylus {
 	gint refcnt;
-	int id;
+	WacomStylusId id;
 	char *name;
 	char *group;
 	int num_buttons;
 	gboolean has_eraser;
-	GArray *paired_ids;
+	GArray *paired_styli; /* [WacomStylus*, ...] */
+	GArray *deprecated_paired_ids; /* [int, ...] */
+	GArray *paired_stylus_ids; /* [WacomStylusId, ...], NULL once parsing is complete */
 	WacomEraserType eraser_type;
 	gboolean has_lens;
 	gboolean has_wheel;
@@ -124,7 +153,7 @@ struct _WacomStylus {
 
 struct _WacomDeviceDatabase {
 	GHashTable *device_ht; /* key = DeviceMatch (str), value = WacomDeviceData * */
-	GHashTable *stylus_ht; /* key = ID (int), value = WacomStylus * */
+	GHashTable *stylus_ht; /* key = WacomStylusId, value = WacomStylus * */
 };
 
 struct _WacomError {
@@ -143,12 +172,13 @@ void libwacom_error_set(WacomError *error, enum WacomErrorCode code, const char
 void libwacom_add_match(WacomDevice *device, WacomMatch *newmatch);
 void libwacom_set_default_match(WacomDevice *device, WacomMatch *newmatch);
 void libwacom_remove_match(WacomDevice *device, WacomMatch *newmatch);
-WacomMatch* libwacom_match_new(const char *name, WacomBusType bus,
+WacomMatch* libwacom_match_new(const char *name, const char *uniq,
+			       WacomBusType bus,
 			       int vendor_id, int product_id);
 
 WacomBusType  bus_from_str (const char *str);
 const char   *bus_to_str   (WacomBusType bus);
-char *make_match_string(const char *name, WacomBusType bus, int vendor_id, int product_id);
+char *make_match_string(const char *name, const char *uniq, WacomBusType bus, int vendor_id, int product_id);
 
 #endif /* _LIBWACOMINT_H_ */
 
