@@ -6,12 +6,12 @@ date=`date +"%Y-%m-%d-%H.%M.%S"`
 builddir="build.$date"
 
 echo "####################################### running test suite"
-meson $builddir
+meson setup $builddir
 ninja -C $builddir test
 
 echo "####################################### running valgrind"
 pushd $builddir > /dev/null
-meson test --setup=valgrind --suite=valgrind
+meson test --setup=valgrind
 popd > /dev/null
 
 echo "####################################### running ubsan"
