@@ -45,7 +45,7 @@ class _Api:
 
     @property
     def basename(self) -> str:
-        return self.name[len(PREFIX) :]
+        return self.name.removeprefix(PREFIX)
 
 
 @dataclass
@@ -58,6 +58,39 @@ class _Enum:
         return self.name.removeprefix("WACOM_").removeprefix("W")
 
 
+class GlibC:
+    _lib = None
+
+    _api_prototypes: List[_Api] = [
+        _Api(name="free", args=(c_void_p,), return_type=None),
+    ]
+
+    @staticmethod
+    def _cdll():
+        # BSD has 7, Linux has 6
+        for libc in ("libc.so.6", "libc.so.7"):
+            try:
+                return ctypes.CDLL(libc, use_errno=True)
+            except OSError:
+                pass
+        raise NotImplementedError("Not implemented for other libc.so")
+
+    @classmethod
+    def _load(cls):
+        cls._lib = cls._cdll()
+        for api in cls._api_prototypes:
+            func = getattr(cls._lib, api.name)
+            func.argtypes = api.args
+            func.restype = api.return_type
+            setattr(cls, api.basename, func)
+
+    @classmethod
+    def instance(cls):
+        if cls._lib is None:
+            cls._load()
+        return cls
+
+
 class LibWacom:
     """
     libwacom.so wrapper. This is a singleton ctypes wrapper into libwacom.so with
@@ -163,8 +196,13 @@ class LibWacom:
         _Api(name="libwacom_get_num_keys", args=(c_void_p,), return_type=c_int),
         _Api(
             name="libwacom_get_supported_styli",
+            args=(c_void_p, ctypes.POINTER(c_int)),
+            return_type=ctypes.POINTER(c_int),
+        ),
+        _Api(
+            name="libwacom_get_styli",
             args=(c_void_p, c_void_p),
-            return_type=c_void_p,
+            return_type=ctypes.POINTER(c_void_p),
         ),
         _Api(name="libwacom_has_ring", args=(c_void_p,), return_type=c_int),
         _Api(name="libwacom_has_ring2", args=(c_void_p,), return_type=c_int),
@@ -180,7 +218,7 @@ class LibWacom:
         _Api(
             name="libwacom_get_status_leds",
             args=(c_void_p, c_void_p),
-            return_type=c_void_p,
+            return_type=ctypes.POINTER(ctypes.c_int),
         ),
         _Api(
             name="libwacom_get_button_led_group",
@@ -207,6 +245,7 @@ class LibWacom:
             return_type=c_void_p,
         ),
         _Api(name="libwacom_stylus_get_id", args=(c_void_p,), return_type=c_int),
+        _Api(name="libwacom_stylus_get_vendor_id", args=(c_void_p,), return_type=c_int),
         _Api(name="libwacom_stylus_get_name", args=(c_void_p,), return_type=c_char_p),
         _Api(
             name="libwacom_stylus_get_paired_ids",
@@ -225,6 +264,11 @@ class LibWacom:
         _Api(
             name="libwacom_stylus_get_eraser_type", args=(c_void_p,), return_type=c_int
         ),
+        _Api(
+            name="libwacom_stylus_get_paired_styli",
+            args=(c_void_p, c_void_p),
+            return_type=ctypes.POINTER(c_void_p),
+        ),
         _Api(
             name="libwacom_print_stylus_description",
             args=(c_int, c_void_p),
@@ -350,6 +394,8 @@ class LibWacom:
         _Enum(name="WACOM_STATUS_LED_RING2", value=2),
         _Enum(name="WACOM_STATUS_LED_TOUCHSTRIP", value=3),
         _Enum(name="WACOM_STATUS_LED_TOUCHSTRIP2", value=4),
+        _Enum(name="WACOM_STATUS_LED_DIAL", value=1),
+        _Enum(name="WACOM_STATUS_LED_DIAL2", value=2),
     ]
 
 
@@ -491,6 +537,102 @@ class WacomBuilder:
         lib.builder_destroy(self.builder)
 
 
+class WacomStylusType(enum.IntEnum):
+    UNKNOWN = 0
+    GENERAL = 1
+    INKING = 2
+    AIRBRUSH = 3
+    CLASSIC = 4
+    MARKER = 5
+    STROKE = 6
+    PUCK = 7
+    THREED = 8
+    MOBILE = 9
+
+
+class WacomEraserType(enum.IntEnum):
+    UNKNOWN = 0
+    NONE = 1
+    INVERT = 2
+    BUTTON = 3
+
+
+class WacomStylus:
+    class Generic(enum.IntEnum):
+        PEN_WITH_ERASER = 0xFFFFF
+        ERASER = 0xFFFFE
+        PEN_NO_ERASER = 0xFFFFD
+
+    def __init__(self, stylus):
+        self.stylus = stylus
+        lib = LibWacom.instance()
+
+        def wrapper(func):
+            return lambda *args, **kwargs: func(self.stylus, *args, **kwargs)
+
+        # Map all device-specifice accessors into respective functions
+        for api in lib._api_prototypes:
+            allowlist = ["stylus"]
+            if any(api.basename.startswith(n) for n in allowlist):
+                denylist = ["stylus_get_paired_styli", "stylus_is_eraser"]
+                if all(not api.basename.startswith(n) for n in denylist):
+                    func = getattr(lib, api.basename)
+                    setattr(self, api.basename.removeprefix("stylus_"), wrapper(func))
+
+    @property
+    def name(self):
+        return self.get_name().decode("utf-8")
+
+    @property
+    def group(self):
+        return self.get_group().decode("utf-8")
+
+    @property
+    def tool_id(self) -> int:
+        return self.get_id()
+
+    @property
+    def vendor_id(self) -> int:
+        return self.get_vendor_id()
+
+    @property
+    def num_buttons(self) -> int:
+        return self.get_num_buttons()
+
+    @property
+    def is_eraser(self) -> bool:
+        lib = LibWacom.instance()
+        return lib.stylus_is_eraser(self.stylus) != 0
+
+    @property
+    def stylus_type(self) -> WacomEraserType:
+        return WacomEraserType(self.get_eraser_type())
+
+    @property
+    def eraser_type(self) -> WacomEraserType:
+        return WacomEraserType(self.get_eraser_type())
+
+    def get_paired_styli(self) -> List["WacomStylus"]:
+        lib = LibWacom.instance()
+        paired = lib.stylus_get_paired_styli(self.stylus, None)
+        styli = [
+            WacomStylus(p)
+            for p in itertools.takewhile(lambda ptr: ptr is not None, paired)
+        ]
+        GlibC.instance().free(paired)
+        return styli
+
+
+class WacomStatusLed(enum.IntEnum):
+    UNAVAILABLE = -1
+    RING = 0
+    RING2 = 1
+    TOUCHSTRIP = 2
+    TOUCHSTRIP2 = 3
+    DIAL = 4
+    DIAL2 = 5
+
+
 class WacomDevice:
     """
     Convenience wrapper to make using libwacom a bit more pythonic.
@@ -538,7 +680,7 @@ class WacomDevice:
         for api in lib._api_prototypes:
             allowlist = ["get_", "is_", "has_"]
             if any(api.basename.startswith(n) for n in allowlist):
-                denylist = ["get_paired_device", "get_matches"]
+                denylist = ["get_paired_device", "get_matches", "get_styli"]
                 if all(not api.basename.startswith(n) for n in denylist):
                     func = getattr(lib, api.basename)
                     setattr(self, api.basename, wrapper(func))
@@ -562,6 +704,15 @@ class WacomDevice:
             for m in itertools.takewhile(lambda ptr: ptr is not None, matches)
         ]
 
+    def get_styli(self) -> List[WacomStylus]:
+        lib = LibWacom.instance()
+        styli = lib.get_styli(self.device, None)
+
+        return [
+            WacomStylus(m)
+            for m in itertools.takewhile(lambda ptr: ptr is not None, styli)
+        ]
+
     def __del__(self):
         if self._destroy_on_del:
             lib = LibWacom.instance()
@@ -661,6 +812,16 @@ class WacomDevice:
     def button_evdev_code(self, button: str) -> int:
         return self.get_button_evdev_code(button.encode("utf-8"))
 
+    def button_led_group(self, button: str) -> List[ButtonFlags]:
+        return self.get_button_led_group(button.encode("utf-8"))
+
+    @property
+    def status_leds(self) -> List["WacomStatusLed"]:
+        nleds = c_int()
+        leds = self.get_status_leds(ctypes.byref(nleds))
+
+        return [WacomStatusLed(l) for l in leds[: nleds.value]]
+
 
 class WacomDatabase:
     """
@@ -716,7 +877,9 @@ class WacomDatabase:
 
     def list_devices(self) -> List[WacomDevice]:
         devices = self.libwacom_list_devices_from_database(self.db, 0)
-        return [
+        devs = [
             WacomDevice(d, destroy=False)
             for d in itertools.takewhile(lambda ptr: ptr is not None, devices)
         ]
+        GlibC.instance().free(devices)
+        return devs
