@@ -2,69 +2,142 @@
 #
 # This file is formatted with ruff format
 
+from configparser import ConfigParser
+from dataclasses import dataclass, field
+
+import ctypes
 import logging
 import pytest
 
-from . import WacomBuilder, WacomBustype, WacomDatabase, WacomDevice
+from . import (
+    WacomBuilder,
+    WacomBustype,
+    WacomDatabase,
+    WacomDevice,
+    WacomEraserType,
+    WacomStatusLed,
+    WacomStylus,
+)
 
 logger = logging.getLogger(__name__)
 
 
-def write_stylus_file(dir):
-    from configparser import ConfigParser
+@dataclass
+class StylusEntry:
+    id: str
+    name: str
+    group: str = "generic-with-eraser"
+    paired_stylus_ids: list[str] = field(default_factory=list)
+    buttons: int = 2
+    axes: str = "Tilt;Pressure;Distance;"
+    stylus_type: str = "General"
+    eraser_type: str | None = None
 
-    config = ConfigParser()
-    config.optionxform = lambda option: option
-    config["0xfffff"] = {
-        "Name": "General Pen",
-        "Group": "generic-with-eraser",
-        "PairedStylusIds": "0xffffe;",
-        "Buttons": "2",
-        "Axes": "Tilt;Pressure;Distance;",
-        "Type": "General",
-    }
+    def add_to_config(self, config: ConfigParser):
+        c = {
+            "Name": self.name,
+            "Group": self.group,
+            "PairedStylusIds": ";".join(self.paired_stylus_ids),
+            "Buttons": f"{self.buttons}",
+            "Axes": self.axes,
+            "Type": self.stylus_type,
+        }
+        if self.eraser_type is not None:
+            c["EraserType"] = self.eraser_type
+        config[self.id] = c
 
-    config["0xffffe"] = {
-        "Name": "General Pen Eraser",
-        "Group": "generic-with-eraser",
-        "PairedStylusIds": "0xfffff;",
-        "EraserType": "Invert",
-        "Buttons": "2",
-        "Axes": "Tilt;Pressure;Distance;",
-        "Type": "General",
-    }
+    @classmethod
+    def generic_pen(cls) -> "StylusEntry":
+        return cls(
+            id="0x0:0xfffff", name="General Pen", paired_stylus_ids=["0x0:0xffffe"]
+        )
 
-    with open(dir / "libwacom.stylus", "w") as fd:
-        config.write(fd, space_around_delimiters=False)
+    @classmethod
+    def generic_eraser(cls) -> "StylusEntry":
+        return StylusEntry(
+            id="0x0:0xffffe",
+            name="General Pen Eraser",
+            paired_stylus_ids=["0x0:0xfffff"],
+            eraser_type="Invert",
+        )
 
 
-def write_tablet_file(filename, devicename, matches):
-    from configparser import ConfigParser
+@dataclass
+class StylusFile:
+    entries: list[StylusEntry]
 
-    config = ConfigParser()
-    config.optionxform = lambda option: option
-    config["Device"] = {
-        "Name": devicename,
-        "DeviceMatch": ";".join(matches),
-        "Width": 9,
-        "Height": 6,
-        "IntegratedIn": "",
-        "Class": "Bamboo",
-        "Layout": "",
-    }
-    config["Features"] = {
-        "Stylus": True,
-        "Reversible": False,
-    }
+    @classmethod
+    def default(cls) -> "StylusFile":
+        return cls(
+            entries=[
+                StylusEntry.generic_pen(),
+                StylusEntry.generic_eraser(),
+            ]
+        )
 
-    with open(filename, "w") as fd:
-        config.write(fd, space_around_delimiters=False)
+    def write_to_dir(self, dir, filename="libwacom.stylus"):
+        config = ConfigParser()
+        config.optionxform = lambda option: option
+
+        for s in self.entries:
+            s.add_to_config(config)
+
+        with open(dir / filename, "w") as fd:
+            config.write(fd, space_around_delimiters=False)
+
+
+@dataclass
+class TabletFile:
+    name: str
+    matches: list[str]
+    width: int = 9
+    height: int = 6
+    integrated_in: str = ""
+    klass: str = "Bamboo"
+    layout: str = ""
+    has_stylus: bool = True
+    is_reversible: bool = False
+    styli: list[str] = field(default_factory=list)
+    # extra additions to the tablet file, e.g. { "Buttons": { "Left" : "A;B;" }
+    extra: dict[str, dict[str, str]] = field(default_factory=dict)
+
+    def write_to(self, filename):
+        config = ConfigParser()
+        config.optionxform = lambda option: option
+
+        cfg = {}
+        cfg["Device"] = {
+            "Name": self.name,
+            "DeviceMatch": ";".join(self.matches),
+            "Width": self.width,
+            "Height": self.height,
+            "IntegratedIn": self.integrated_in,
+            "Class": self.klass,
+            "Layout": self.layout,
+        }
+        if self.styli:
+            cfg["Device"]["Styli"] = ";".join(self.styli)
+
+        cfg["Features"] = {
+            "Stylus": self.has_stylus,
+            "Reversible": self.is_reversible,
+        }
+
+        cfg = cfg | self.extra
+
+        for k, v in cfg.items():
+            config[k] = v
+
+        with open(filename, "w") as fd:
+            config.write(fd, space_around_delimiters=False)
 
 
 @pytest.fixture()
 def custom_datadir(tmp_path):
-    write_stylus_file(tmp_path)
-    write_tablet_file(tmp_path / "generic.tablet", "Generic", ["generic"])
+    StylusFile.default().write_to_dir(tmp_path)
+    TabletFile(name="Generic", matches=["generic"]).write_to(
+        tmp_path / "generic.tablet"
+    )
     return tmp_path
 
 
@@ -210,6 +283,32 @@ def test_isdv4_4800(db):
     assert device.num_buttons == 0
 
 
+@pytest.mark.parametrize(
+    "usbid,expected",
+    [
+        [(0x256C, 0x0067), [WacomStylus.Generic.PEN_NO_ERASER]],
+        [
+            (0x04F3, 0x264C),
+            [
+                WacomStylus.Generic.PEN_WITH_ERASER,
+                WacomStylus.Generic.ERASER,
+                WacomStylus.Generic.PEN_NO_ERASER,
+            ],
+        ],
+    ],
+)
+def test_generic_pens(db, usbid, expected):
+    # Inspiroy 2 has a generic-pen-no-eraser
+    device = db.new_from_usbid(*usbid)
+    assert device is not None
+
+    nstyli = ctypes.c_int()
+    styli = device.get_supported_styli(ctypes.byref(nstyli))
+    s = [WacomStylus.Generic(id) for id in styli[: nstyli.value]]
+
+    assert sorted(s) == sorted(expected)
+
+
 @pytest.mark.parametrize(
     "bus,vid,pid",
     [
@@ -287,15 +386,19 @@ def test_exact_matches(custom_datadir):
 
     # A device match with uniq but no name
     matches = ["usb|1234|5678||uniqval"]
-    write_tablet_file(custom_datadir / "uniq.tablet", "UniqOnly", matches)
+    TabletFile(name="UniqOnly", matches=matches).write_to(
+        custom_datadir / "uniq.tablet"
+    )
 
     # A device match with a name but no uniq
     matches = ["usb|1234|5678|nameval"]
-    write_tablet_file(custom_datadir / "name.tablet", "NameOnly", matches)
+    TabletFile(name="NameOnly", matches=matches).write_to(
+        custom_datadir / "name.tablet"
+    )
 
     # A device match with both
     matches = ["usb|1234|5678|nameval|uniqval"]
-    write_tablet_file(custom_datadir / "both.tablet", "Both", matches)
+    TabletFile(name="Both", matches=matches).write_to(custom_datadir / "both.tablet")
 
     db = WacomDatabase(path=custom_datadir)
 
@@ -322,11 +425,15 @@ def test_prefer_uniq_over_name(custom_datadir):
 
     # A device match with uniq but no name
     matches = ["usb|1234|5678||uniqval"]
-    write_tablet_file(custom_datadir / "uniq.tablet", "UniqOnly", matches)
+    TabletFile(name="UniqOnly", matches=matches).write_to(
+        custom_datadir / "uniq.tablet"
+    )
 
     # A device match with a name but no uniq
     matches = ["usb|1234|5678|nameval"]
-    write_tablet_file(custom_datadir / "name.tablet", "NameOnly", matches)
+    TabletFile(name="NameOnly", matches=matches).write_to(
+        custom_datadir / "name.tablet"
+    )
 
     db = WacomDatabase(path=custom_datadir)
 
@@ -359,7 +466,7 @@ def test_dont_ignore_exact_matches(custom_datadir):
 
     # A device match with both
     matches = ["usb|1234|5678|nameval|uniqval"]
-    write_tablet_file(custom_datadir / "both.tablet", "Both", matches)
+    TabletFile(name="Both", matches=matches).write_to(custom_datadir / "both.tablet")
 
     db = WacomDatabase(path=custom_datadir)
 
@@ -465,3 +572,232 @@ def test_new_from_path_unknown_device(db, fallback):
         assert dev.name == name
         assert dev.vendor_id == 0
         assert dev.product_id == 0
+
+
+@pytest.mark.parametrize(
+    "feature",
+    ("Ring", "Touchstrip", "Dial"),
+)
+@pytest.mark.parametrize("count", (1, 2))
+def test_button_modeswitch(custom_datadir, feature, count):
+    USBID = (0x1234, 0x5678)
+
+    # sigh, Touchstrip but StripsNumModes...
+    num_mode_key = f"{feature}s" if feature != "Touchstrip" else "Strip"
+
+    extra = {
+        "Buttons": {
+            "Left": "A;B;C;",
+            "Right": "D;E;F;",
+            feature: "A",
+        },
+        "Features": {
+            num_mode_key: 4,
+        },
+    }
+    if count > 1:
+        extra["Buttons"][f"{feature}2"] = "D"
+
+    TabletFile(
+        name="some tablet",
+        matches=[f"usb|{USBID[0]:04x}|{USBID[1]:04x}"],
+        extra=extra,
+    ).write_to(custom_datadir / "led.tablet")
+
+    db = WacomDatabase(path=custom_datadir)
+    builder = WacomBuilder.create(usbid=USBID)
+    device = db.new_from_builder(builder)
+    assert device is not None
+
+    expected_flag = {
+        "Ring": WacomDevice.ButtonFlags.RING_MODESWITCH,
+        "Touchstrip": WacomDevice.ButtonFlags.TOUCHSTRIP_MODESWITCH,
+        "Dial": WacomDevice.ButtonFlags.DIAL_MODESWITCH,
+    }[feature]
+
+    flags = device.button_flags("A")
+    assert expected_flag in flags
+
+    for b in "BCDEF":
+        flags = device.button_flags(b)
+        assert expected_flag not in flags
+
+    expected_flag = {
+        "Ring": WacomDevice.ButtonFlags.RING2_MODESWITCH,
+        "Touchstrip": WacomDevice.ButtonFlags.TOUCHSTRIP2_MODESWITCH,
+        "Dial": WacomDevice.ButtonFlags.DIAL2_MODESWITCH,
+    }[feature]
+
+    flags = device.button_flags("D")
+    if count > 1:
+        assert expected_flag in flags
+    else:
+        assert expected_flag not in flags
+
+    for b in "ABCEF":
+        flags = device.button_flags(b)
+        assert expected_flag not in flags
+
+
+@pytest.mark.parametrize(
+    "feature",
+    ("Ring", "Touchstrip", "Dial"),
+)
+@pytest.mark.parametrize("count", (1, 2))
+def test_status_leds(custom_datadir, feature, count):
+    USBID = (0x1234, 0x5678)
+
+    # sigh, Touchstrip but StripsNumModes...
+    num_mode_key = f"{feature}s" if feature != "Touchstrip" else "Strip"
+
+    extra = {
+        "Buttons": {
+            "Left": "A;B;C;",
+            "Right": "D;E;F;",
+            feature: "A",
+        },
+        "Features": {
+            "StatusLEDs": f"{feature};{feature}2" if count > 1 else f"{feature}",
+            num_mode_key: 4,
+        },
+    }
+    if count > 1:
+        extra["Buttons"][f"{feature}2"] = "D"
+
+    TabletFile(
+        name="some tablet",
+        matches=[f"usb|{USBID[0]:04x}|{USBID[1]:04x}"],
+        extra=extra,
+    ).write_to(custom_datadir / "led.tablet")
+
+    db = WacomDatabase(path=custom_datadir)
+    builder = WacomBuilder.create(usbid=USBID)
+    device = db.new_from_builder(builder)
+    assert device is not None
+
+    expected = [
+        {
+            "Ring": WacomStatusLed.RING,
+            "Touchstrip": WacomStatusLed.TOUCHSTRIP,
+            "Dial": WacomStatusLed.DIAL,
+        }[feature]
+    ]
+
+    if count > 1:
+        expected.append(
+            {
+                "Ring": WacomStatusLed.RING2,
+                "Touchstrip": WacomStatusLed.TOUCHSTRIP2,
+                "Dial": WacomStatusLed.DIAL2,
+            }[feature]
+        )
+
+    leds = device.status_leds
+    assert sorted(leds) == sorted(expected)
+
+    led_group = device.button_led_group("A")
+    assert led_group == 0
+
+    led_group = device.button_led_group("D")
+    if count > 1:
+        assert led_group == 1
+    else:
+        assert led_group == -1
+
+    for b in "BCEF":
+        led_group = device.button_led_group(b)
+        assert led_group == -1
+
+
+def test_nonwacom_stylus_ids(tmp_path):
+    styli = StylusFile.default()
+    s1 = StylusEntry(
+        id="0x1234:0xabcd",
+        name="ABC Pen",
+        group="notwacom",
+        paired_stylus_ids=["0x1234:0x9876"],
+    )
+    s2 = StylusEntry(
+        id="0x1234:0x9876",
+        name="9876 Pen",
+        group="notwacom",
+        paired_stylus_ids=["0x1234:0xabcd"],
+        eraser_type="Invert",
+    )
+    styli.entries.append(s1)
+    styli.entries.append(s2)
+    styli.write_to_dir(tmp_path)
+
+    # matches our nonwacom group
+    TabletFile(
+        name="ABC Group Tablet",
+        matches=["usb|9999|abcd"],
+        styli=["@notwacom"],
+    ).write_to(tmp_path / "group.tablet")
+
+    # matches one nonwacom styli and the default generic ones
+    TabletFile(
+        name="ABC Tablet",
+        matches=["usb|8888|abcd"],
+        styli=[s2.id, "@generic-with-eraser"],
+    ).write_to(tmp_path / "abc.tablet")
+
+    db = WacomDatabase(path=tmp_path)
+
+    builder = WacomBuilder.create(usbid=(0x9999, 0xABCD))
+    device = db.new_from_builder(builder)
+    assert device is not None
+    assert device.name == "ABC Group Tablet"
+    styli = device.get_styli()
+    assert len(styli) == 2
+    assert styli[0].vendor_id == 0x1234
+    assert styli[1].vendor_id == 0x1234
+    # Order of styli is undefined
+    assert styli[0].tool_id == 0xABCD or styli[1].tool_id == 0xABCD
+    assert styli[0].tool_id == 0x9876 or styli[1].tool_id == 0x9876
+    assert sum(s.is_eraser for s in styli) == 1
+    for s in filter(lambda s: s.is_eraser, styli):
+        assert s.eraser_type == WacomEraserType.INVERT
+
+    assert sum(s.is_eraser is True for s in styli) == 1
+
+    paired0 = styli[0].get_paired_styli()
+    paired1 = styli[1].get_paired_styli()
+    assert len(paired0) == 1
+    assert len(paired1) == 1
+    assert paired0[0].vendor_id == styli[1].vendor_id
+    assert paired0[0].tool_id != styli[1].vendor_id
+    assert paired1[0].vendor_id == styli[0].vendor_id
+    assert paired1[0].tool_id != styli[0].vendor_id
+
+    builder = WacomBuilder.create(usbid=(0x8888, 0xABCD))
+    device = db.new_from_builder(builder)
+    assert device is not None
+    assert device.name == "ABC Tablet"
+    styli = device.get_styli()
+    assert len(styli) == 3  # 1 non-wacom, 2 generic ones
+    # Order of styli is undefined
+    assert sum(s.vendor_id == 0x1234 and s.tool_id == 0x9876 for s in styli) == 1
+    assert sum(s.vendor_id == 0 and s.tool_id == 0xFFFFE for s in styli) == 1
+    assert sum(s.vendor_id == 0 and s.tool_id == 0xFFFFF for s in styli) == 1
+
+
+def test_load_xdg_config_home(monkeypatch, tmp_path, custom_datadir):
+    monkeypatch.setenv("XDG_CONFIG_HOME", str(tmp_path.absolute()))
+
+    xdg = tmp_path / "libwacom"
+    xdg.mkdir()
+
+    usbid = (0x1234, 0x5678)
+    matches = [f"usb|{usbid[0]:04x}|{usbid[1]:04x}"]
+    TabletFile(name="XDGTablet", matches=matches).write_to(xdg / "uniq.tablet")
+
+    StylusFile.default().write_to_dir(xdg)
+
+    # This should load from system *and* XDG. system files could
+    # interfere with our test or may not exist but unfortunately we can't
+    # chroot for the test. it should be good enough this way anyway.
+    db = WacomDatabase()
+    builder = WacomBuilder.create(usbid=usbid)
+    device = db.new_from_builder(builder)
+    assert device is not None and device.name == "XDGTablet"
