@@ -1,4 +1,4 @@
-#! /usr/bin/env python
+#! /usr/bin/env python3
 #
 # Copyright (c) 2013 Red Hat, Inc.
 #
@@ -18,19 +18,26 @@
 # Author: Joaquim Rocha <jrocha@redhat.com>
 #
 
+import logging
 import sys
+from pathlib import Path
 from argparse import ArgumentParser
 from xml.etree import ElementTree as ET
 
+logging.basicConfig(level=logging.INFO)
+logger = logging.getLogger("clean-svg")
+logger.setLevel(logging.INFO)
+
 NAMESPACE = "http://www.w3.org/2000/svg"
 BRACKETS_NAMESPACE = "{" + NAMESPACE + "}"
 
 
 def human_round(number):
     """
-    Round to closest .5
+    Round to closest .5, keep integer values
     """
-    return round(number * 2) / 2.0
+    v = round(number * 2) / 2.0
+    return int(v) if v == int(v) else v
 
 
 def traverse_and_clean(node):
@@ -41,9 +48,6 @@ def traverse_and_clean(node):
     for key in list(node.attrib.keys()):
         if key.startswith("{"):
             del node.attrib[key]
-    if node.tag == "g" and "id" in node.attrib:
-        apply_id_and_class_from_group(node)
-        del node.attrib["id"]
     if "style" in node.attrib:
         if node.tag == "text":
             node.attrib["style"] = "text-anchor:start;"
@@ -82,7 +86,7 @@ def round_if_number(value):
 def remove_non_svg_nodes_and_strip_namespace(root):
     if root.tag.startswith(BRACKETS_NAMESPACE):
         root.tag = root.tag[len(BRACKETS_NAMESPACE) :]
-    for elem in root:
+    for elem in list(root):
         if (
             not elem.tag.startswith(BRACKETS_NAMESPACE)
             or elem.tag == BRACKETS_NAMESPACE + "metadata"
@@ -261,7 +265,7 @@ def apply_id_and_class_from_group(group_node):
     _id = group_node.attrib.get("id")
     if _id is None:
         return
-    for child in group_node.getchildren():
+    for child in group_node:
         if child.tag == "rect" or child.tag == "circle":
             if button_assigned:
                 continue
@@ -302,23 +306,62 @@ def clean_svg(root, tabletname):
 if __name__ == "__main__":
     parser = ArgumentParser(description="Clean SVG files for libwacom")
     parser.add_argument(
-        "filename", nargs=1, type=str, help="SVG file to clean", metavar="FILE"
+        "--debug",
+        action="store_true",
+        default=False,
+        help="Print debug logging output",
+    )
+    parser.add_argument(
+        "--ignore-missing",
+        action="store_true",
+        default=False,
+        help="Ignore .tablet files without a Layout",
+    )
+    parser.add_argument(
+        "filename",
+        type=str,
+        help="SVG file to clean or .tablet file with a Layout=xyz.svg line",
+        metavar="FILE",
     )
     parser.add_argument(
         "tabletname",
-        nargs=1,
         type=str,
-        help="The name of the tablet",
+        nargs="?",
+        help="The name of the tablet, can be omitted if FILE is a .tablet file",
         metavar="TABLET_NAME",
     )
     args = parser.parse_args()
 
+    if args.debug:
+        logger.setLevel(logging.DEBUG)
+
+    svgfile = args.filename
+    tabletname = args.tabletname
+    if args.filename.endswith(".tablet"):
+        import configparser
+
+        config = configparser.ConfigParser()
+        config.read(args.filename)
+        try:
+            svgname = config["Device"]["Layout"]
+            if not svgname:
+                raise KeyError("Empty layout line")
+        except KeyError:
+            print(
+                f"{args.filename} does not specify a layout, skipping", file=sys.stderr
+            )
+            sys.exit(0 if args.ignore_missing else 77)
+        svgfile = Path(args.filename).parent / "layouts" / svgname
+        tabletname = config["Device"]["Name"]
+    else:
+        assert tabletname, "Argument TABLET_NAME must be provided for SVG files"
+
     ET.register_namespace("", NAMESPACE)
     try:
-        tree = ET.parse(args.filename[0])
+        tree = ET.parse(svgfile)
     except Exception as e:
         sys.stderr.write(str(e) + "\n")
         sys.exit(1)
     root = tree.getroot()
-    clean_svg(root, args.tabletname[0])
+    clean_svg(root, tabletname)
     print(to_string(root))
